@@ -1,37 +1,45 @@
-/*  This file is part of the Vc library.
+/*  This file is part of the Vc library. {{{
+Copyright © 2009-2015 Matthias Kretz <kretz@kde.org>
 
-    Copyright (C) 2009-2012 Matthias Kretz <kretz@kde.org>
+Redistribution and use in source and binary forms, with or without
+modification, are permitted provided that the following conditions are met:
+    * Redistributions of source code must retain the above copyright
+      notice, this list of conditions and the following disclaimer.
+    * Redistributions in binary form must reproduce the above copyright
+      notice, this list of conditions and the following disclaimer in the
+      documentation and/or other materials provided with the distribution.
+    * Neither the names of contributing organizations nor the
+      names of its contributors may be used to endorse or promote products
+      derived from this software without specific prior written permission.
 
-    Vc is free software: you can redistribute it and/or modify
-    it under the terms of the GNU Lesser General Public License as
-    published by the Free Software Foundation, either version 3 of
-    the License, or (at your option) any later version.
+THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
+DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
+(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
+ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
-    Vc is distributed in the hope that it will be useful, but
-    WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU Lesser General Public License for more details.
+}}}*/
 
-    You should have received a copy of the GNU Lesser General Public
-    License along with Vc.  If not, see <http://www.gnu.org/licenses/>.
-
-*/
-
-#ifndef VC_COMMON_MEMORY_H
-#define VC_COMMON_MEMORY_H
+#ifndef VC_COMMON_MEMORY_H_
+#define VC_COMMON_MEMORY_H_
 
 #include "memorybase.h"
 #include <assert.h>
 #include <algorithm>
 #include <cstring>
 #include <cstddef>
+#include <initializer_list>
 #include "memoryfwd.h"
+#include "malloc.h"
 #include "macros.h"
 
-/*OUTER_NAMESPACE_BEGIN*/
-namespace Vc
+namespace Vc_VERSIONED_NAMESPACE
 {
-
 /**
  * Allocates memory on the Heap with alignment and padding suitable for vectorized access.
  *
@@ -64,9 +72,9 @@ namespace Vc
  * \headerfile memory.h <Vc/Memory>
  */
 template<typename T, Vc::MallocAlignment A>
-Vc_ALWAYS_INLINE_L T *Vc_ALWAYS_INLINE_R malloc(size_t n)
+Vc_ALWAYS_INLINE T *malloc(size_t n)
 {
-    return static_cast<T *>(Internal::Helper::malloc<A>(n * sizeof(T)));
+    return static_cast<T *>(Common::malloc<A>(n * sizeof(T)));
 }
 
 /**
@@ -93,9 +101,11 @@ Vc_ALWAYS_INLINE_L T *Vc_ALWAYS_INLINE_R malloc(size_t n)
 template<typename T>
 Vc_ALWAYS_INLINE void free(T *p)
 {
-    Internal::Helper::free(p);
+    Common::free(p);
 }
 
+namespace Common
+{
 template<typename V, size_t Size> struct _MemorySizeCalculation
 {
     enum AlignmentCalculations {
@@ -117,28 +127,42 @@ template<typename V, size_t Size> struct _MemorySizeCalculation
  * \param Size1 Number of rows
  * \param Size2 Number of columns
  */
-template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAlignedBaseT<V>, public MemoryBase<V, Memory<V, Size1, Size2>, 2, Memory<V, Size2> >
-{
-    public:
-        typedef typename V::EntryType EntryType;
-    private:
-        typedef MemoryBase<V, Memory<V, Size1, Size2>, 2, Memory<V, Size2> > Base;
-            friend class MemoryBase<V, Memory<V, Size1, Size2>, 2, Memory<V, Size2> >;
-            friend class MemoryDimensionBase<V, Memory<V, Size1, Size2>, 2, Memory<V, Size2> >;
-            enum InternalConstants {
-                PaddedSize2 = _MemorySizeCalculation<V, Size2>::PaddedSize
-            };
-#if defined(VC_ICC) && defined(_WIN32)
-            __declspec(align(__alignof(VectorAlignedBaseT<V>)))
-#elif defined(VC_CLANG)
-            __attribute__((aligned(__alignof(VectorAlignedBaseT<V>))))
-#elif defined(VC_MSVC)
-	    VectorAlignedBaseT<V> _force_alignment;
-            // __declspec(align(#)) accepts only numbers not __alignof nor just VectorAlignment
-	    // by putting VectorAlignedBaseT<V> here _force_alignment is aligned correctly.
-	    // the downside is that there's a lot of padding before m_mem (32 Bytes with SSE) :(
+template <typename V, size_t Size1, size_t Size2, bool InitPadding>
+#ifdef Vc_RECURSIVE_MEMORY
+class Memory : public MemoryBase<V, Memory<V, Size1, Size2, InitPadding>, 2,
+                                 Memory<V, Size2, 0, InitPadding>>
+#else
+class Memory : public AlignedBase<V::MemoryAlignment>,
+               public MemoryBase<V, Memory<V, Size1, Size2, InitPadding>, 2,
+                                 Memory<V, Size2, 0, false>>
 #endif
-            EntryType m_mem[Size1][PaddedSize2];
+{
+public:
+    typedef typename V::EntryType EntryType;
+
+private:
+#ifdef Vc_RECURSIVE_MEMORY
+    using RowMemory = Memory<V, Size2, 0, InitPadding>;
+#else
+    using RowMemory = Memory<V, Size2, 0, false>;
+#endif
+    typedef MemoryBase<V, Memory<V, Size1, Size2, InitPadding>, 2, RowMemory> Base;
+    friend class MemoryBase<V, Memory<V, Size1, Size2, InitPadding>, 2, RowMemory>;
+    friend class MemoryDimensionBase<V, Memory<V, Size1, Size2, InitPadding>, 2,
+                                     RowMemory>;
+    enum : size_t {
+        Alignment = V::MemoryAlignment,
+        PaddedSize2 = _MemorySizeCalculation<V, Size2>::PaddedSize
+    };
+    alignas(static_cast<size_t>(Alignment))  // GCC complains about 'is not an
+                                             // integer constant' unless the
+                                             // static_cast is present
+#ifdef Vc_RECURSIVE_MEMORY
+        RowMemory m_mem[Size1];
+#else
+        EntryType m_mem[Size1][PaddedSize2];
+#endif
+
         public:
             using Base::vector;
             enum Constants {
@@ -146,12 +170,26 @@ template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAli
                 VectorsCount = PaddedSize2 / V::Size
             };
 
+#ifdef Vc_RECURSIVE_MEMORY
+            Memory() = default;
+#else
+            Memory()
+            {
+                if (InitPadding) {
+                    if (Size1 > 32)
+                    for (size_t i = 0; i < Size1; ++i) {
+                        V::Zero().store(&m_mem[i][PaddedSize2 - V::Size], Vc::Streaming);
+                    }
+                }
+            }
+#endif
+
             /**
              * \return the number of rows in the array.
              *
              * \note This function can be eliminated by an optimizing compiler.
              */
-            _VC_CONSTEXPR size_t rowsCount() const { return RowCount; }
+            static constexpr size_t rowsCount() { return RowCount; }
             /**
              * \return the number of scalar entries in the whole array.
              *
@@ -160,13 +198,13 @@ template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAli
              *
              * \note This function can be optimized into a compile-time constant.
              */
-            _VC_CONSTEXPR size_t entriesCount() const { return Size1 * Size2; }
+            static constexpr size_t entriesCount() { return Size1 * Size2; }
             /**
              * \return the number of vectors in the whole array.
              *
              * \note This function can be optimized into a compile-time constant.
              */
-            _VC_CONSTEXPR size_t vectorsCount() const { return VectorsCount * Size1; }
+            static constexpr size_t vectorsCount() { return VectorsCount * Size1; }
 
             /**
              * Copies the data from a different object.
@@ -180,12 +218,12 @@ template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAli
             template<typename Parent, typename RM>
             Vc_ALWAYS_INLINE Memory &operator=(const MemoryBase<V, Parent, 2, RM> &rhs) {
                 assert(vectorsCount() == rhs.vectorsCount());
-                Internal::copyVectors(*this, rhs);
+                Detail::copyVectors(*this, rhs);
                 return *this;
             }
 
             Vc_ALWAYS_INLINE Memory &operator=(const Memory &rhs) {
-                Internal::copyVectors(*this, rhs);
+                Detail::copyVectors(*this, rhs);
                 return *this;
             }
 
@@ -202,11 +240,7 @@ template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAli
                 }
                 return *this;
             }
-    }
-#if defined(VC_ICC) && VC_ICC < 20120212 && !defined(_WIN32)
-    __attribute__((__aligned__(__alignof(VectorAlignedBaseT<V>))))
-#endif
-    ;
+};
 
     /**
      * A helper class to simplify usage of correctly aligned and padded memory, allowing both vector and
@@ -251,32 +285,31 @@ template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAli
      * \ingroup Utilities
      * \headerfile memory.h <Vc/Memory>
      */
-    template<typename V, size_t Size> class Memory<V, Size, 0u> : public VectorAlignedBaseT<V>, public MemoryBase<V, Memory<V, Size, 0u>, 1, void>
+template <typename V, size_t Size, bool InitPadding>
+class Memory<V, Size, 0u, InitPadding> :
+#ifndef Vc_RECURSIVE_MEMORY
+    public AlignedBase<V::MemoryAlignment>,
+#endif
+    public MemoryBase<V, Memory<V, Size, 0u, InitPadding>, 1, void>
     {
         public:
             typedef typename V::EntryType EntryType;
         private:
-            typedef MemoryBase<V, Memory<V, Size, 0u>, 1, void> Base;
-            friend class MemoryBase<V, Memory<V, Size, 0u>, 1, void>;
-            friend class MemoryDimensionBase<V, Memory<V, Size, 0u>, 1, void>;
-            enum InternalConstants {
-                Alignment = V::Size,
-                AlignmentMask = Alignment - 1,
-                MaskedSize = Size & AlignmentMask,
-                Padding = Alignment - MaskedSize,
+            typedef MemoryBase<V, Memory<V, Size, 0u, InitPadding>, 1, void> Base;
+            friend class MemoryBase<V, Memory<V, Size, 0u, InitPadding>, 1, void>;
+            friend class MemoryDimensionBase<V, Memory<V, Size, 0u, InitPadding>, 1, void>;
+            enum : size_t {
+                Alignment = V::MemoryAlignment,     // in Bytes
+                MaskedSize = Size & (V::Size - 1),  // the fraction of Size that exceeds
+                                                    // an integral multiple of V::Size
+                Padding = V::Size - MaskedSize,
                 PaddedSize = MaskedSize == 0 ? Size : Size + Padding
             };
-#if defined(VC_ICC) && defined(_WIN32)
-            __declspec(align(__alignof(VectorAlignedBaseT<V>)))
-#elif defined(VC_CLANG)
-            __attribute__((aligned(__alignof(VectorAlignedBaseT<V>))))
-#elif defined(VC_MSVC)
-	    VectorAlignedBaseT<V> _force_alignment;
-            // __declspec(align(#)) accepts only numbers not __alignof nor just VectorAlignment
-	    // by putting VectorAlignedBaseT<V> here _force_alignment is aligned correctly.
-	    // the downside is that there's a lot of padding before m_mem (32 Bytes with SSE) :(
-#endif
-            EntryType m_mem[PaddedSize];
+            alignas(static_cast<size_t>(Alignment))  // GCC complains about 'is not an
+                                                     // integer constant' unless the
+                                                     // static_cast is present
+                EntryType m_mem[PaddedSize];
+
         public:
             using Base::vector;
             enum Constants {
@@ -284,6 +317,20 @@ template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAli
                 VectorsCount = PaddedSize / V::Size
             };
 
+            Memory()
+            {
+                if (InitPadding) {
+                    Base::lastVector() = V::Zero();
+                }
+            }
+
+            Memory(std::initializer_list<EntryType> init)
+            {
+                Vc_ASSERT(init.size() <= Size);
+                Base::lastVector() = V::Zero();
+                std::copy(init.begin(), init.end(), &m_mem[0]);
+            }
+
             /**
              * Wrap existing data with the Memory convenience class.
              *
@@ -306,14 +353,14 @@ template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAli
              * (not too early/not leaked). This function simply adds convenience functions to \em
              * access the memory.
              */
-            static Vc_ALWAYS_INLINE Vc_CONST Memory<V, Size, 0u> &fromRawData(EntryType *ptr)
+            static Vc_ALWAYS_INLINE Vc_CONST Memory<V, Size, 0u, false> &fromRawData(EntryType *ptr)
             {
                 // DANGER! This placement new has to use the right address. If the compiler decides
                 // RowMemory requires padding before the actual data then the address has to be adjusted
                 // accordingly
                 char *addr = reinterpret_cast<char *>(ptr);
-                typedef Memory<V, Size, 0u> MM;
-                addr -= VC_OFFSETOF(MM, m_mem);
+                typedef Memory<V, Size, 0u, false> MM;
+                addr -= Vc_OFFSETOF(MM, m_mem);
                 return *new(addr) MM;
             }
 
@@ -322,42 +369,36 @@ template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAli
              *
              * \note This function can be optimized into a compile-time constant.
              */
-            _VC_CONSTEXPR size_t entriesCount() const { return EntriesCount; }
+            static constexpr size_t entriesCount() { return EntriesCount; }
 
             /**
              * \return the number of vectors in the whole array.
              *
              * \note This function can be optimized into a compile-time constant.
              */
-            _VC_CONSTEXPR size_t vectorsCount() const { return VectorsCount; }
-
-#ifdef VC_CXX11
-            Vc_ALWAYS_INLINE Memory() = default;
-#else
-            Vc_ALWAYS_INLINE Memory() {}
-#endif
+            static constexpr size_t vectorsCount() { return VectorsCount; }
 
             inline Memory(const Memory &rhs)
             {
-                Internal::copyVectors(*this, rhs);
+                Detail::copyVectors(*this, rhs);
             }
 
             template <size_t S> inline Memory(const Memory<V, S> &rhs)
             {
                 assert(vectorsCount() == rhs.vectorsCount());
-                Internal::copyVectors(*this, rhs);
+                Detail::copyVectors(*this, rhs);
             }
 
             inline Memory &operator=(const Memory &rhs)
             {
-                Internal::copyVectors(*this, rhs);
+                Detail::copyVectors(*this, rhs);
                 return *this;
             }
 
             template <size_t S> inline Memory &operator=(const Memory<V, S> &rhs)
             {
                 assert(vectorsCount() == rhs.vectorsCount());
-                Internal::copyVectors(*this, rhs);
+                Detail::copyVectors(*this, rhs);
                 return *this;
             }
 
@@ -371,11 +412,7 @@ template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAli
                 }
                 return *this;
             }
-    }
-#if defined(VC_ICC) && VC_ICC < 20120212 && !defined(_WIN32)
-    __attribute__((__aligned__(__alignof(VectorAlignedBaseT<V>)) ))
-#endif
-    ;
+    };
 
     /**
      * A helper class that is very similar to Memory<V, Size> but with dynamically allocated memory and
@@ -417,7 +454,7 @@ template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAli
      * \ingroup Utilities
      * \headerfile memory.h <Vc/Memory>
      */
-    template<typename V> class Memory<V, 0u, 0u> : public MemoryBase<V, Memory<V, 0u, 0u>, 1, void>
+    template<typename V> class Memory<V, 0u, 0u, true> : public MemoryBase<V, Memory<V, 0u, 0u, true>, 1, void>
     {
         public:
             typedef typename V::EntryType EntryType;
@@ -453,6 +490,7 @@ template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAli
             m_mem(Vc::malloc<EntryType, Vc::AlignOnVector>(m_vectorsCount))
         {
             m_vectorsCount /= V::Size;
+            Base::lastVector() = V::Zero();
         }
 
         /**
@@ -468,7 +506,7 @@ template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAli
             m_vectorsCount(rhs.vectorsCount()),
             m_mem(Vc::malloc<EntryType, Vc::AlignOnVector>(m_vectorsCount * V::Size))
         {
-            Internal::copyVectors(*this, rhs);
+            Detail::copyVectors(*this, rhs);
         }
 
         /**
@@ -483,7 +521,7 @@ template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAli
             m_vectorsCount(rhs.vectorsCount()),
             m_mem(Vc::malloc<EntryType, Vc::AlignOnVector>(m_vectorsCount * V::Size))
         {
-            Internal::copyVectors(*this, rhs);
+            Detail::copyVectors(*this, rhs);
         }
 
         /**
@@ -527,13 +565,13 @@ template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAli
         template<typename Parent, typename RM>
         Vc_ALWAYS_INLINE Memory &operator=(const MemoryBase<V, Parent, 1, RM> &rhs) {
             assert(vectorsCount() == rhs.vectorsCount());
-            Internal::copyVectors(*this, rhs);
+            Detail::copyVectors(*this, rhs);
             return *this;
         }
 
         Vc_ALWAYS_INLINE Memory &operator=(const Memory &rhs) {
             assert(vectorsCount() == rhs.vectorsCount());
-            Internal::copyVectors(*this, rhs);
+            Detail::copyVectors(*this, rhs);
             return *this;
         }
 
@@ -564,7 +602,7 @@ template<typename V, size_t Size1, size_t Size2> class Memory : public VectorAli
  */
 Vc_ALWAYS_INLINE void prefetchForOneRead(const void *addr)
 {
-    Internal::Helper::prefetchForOneRead(addr);
+    Vc::Detail::prefetchForOneRead(addr, VectorAbi::Best<float>());
 }
 
 /**
@@ -581,7 +619,7 @@ Vc_ALWAYS_INLINE void prefetchForOneRead(const void *addr)
  */
 Vc_ALWAYS_INLINE void prefetchForModify(const void *addr)
 {
-    Internal::Helper::prefetchForModify(addr);
+    Vc::Detail::prefetchForModify(addr, VectorAbi::Best<float>());
 }
 
 /**
@@ -596,7 +634,7 @@ Vc_ALWAYS_INLINE void prefetchForModify(const void *addr)
  */
 Vc_ALWAYS_INLINE void prefetchClose(const void *addr)
 {
-    Internal::Helper::prefetchClose(addr);
+    Vc::Detail::prefetchClose(addr, VectorAbi::Best<float>());
 }
 
 /**
@@ -611,7 +649,7 @@ Vc_ALWAYS_INLINE void prefetchClose(const void *addr)
  */
 Vc_ALWAYS_INLINE void prefetchMid(const void *addr)
 {
-    Internal::Helper::prefetchMid(addr);
+    Vc::Detail::prefetchMid(addr, VectorAbi::Best<float>());
 }
 
 /**
@@ -626,17 +664,21 @@ Vc_ALWAYS_INLINE void prefetchMid(const void *addr)
  */
 Vc_ALWAYS_INLINE void prefetchFar(const void *addr)
 {
-    Internal::Helper::prefetchFar(addr);
+    Vc::Detail::prefetchFar(addr, VectorAbi::Best<float>());
 }
+}  // namespace Common
 
-} // namespace Vc
-/*OUTER_NAMESPACE_END*/
+using Common::Memory;
+using Common::prefetchForOneRead;
+using Common::prefetchForModify;
+using Common::prefetchClose;
+using Common::prefetchMid;
+using Common::prefetchFar;
+}  // namespace Vc
 
 namespace std
 {
     template<typename V> Vc_ALWAYS_INLINE void swap(Vc::Memory<V> &a, Vc::Memory<V> &b) { a.swap(b); }
 } // namespace std
 
-#include "undomacros.h"
-
-#endif // VC_COMMON_MEMORY_H
+#endif // VC_COMMON_MEMORY_H_
